#!/bin/python3
# ./generate-images.py --outdir=../2019-02-18_imgout --cleardir --count 80

import os
import gc
import cv2
import random
import argparse
import numpy as np
import imgaug as ia
from PIL import Image
from PIL import ImageOps
from PIL import ImageFont
from PIL import ImageDraw
from imgaug import augmenters as iaa

import cProfile
import pstats

# import tensorflow as tf
# from object_detection.utils import dataset_util

# Yay globals
isolationsPath = "../isolations/"
backgrounds = [
    os.path.join(isolationsPath, "bricks", x)
    for x in os.listdir(isolationsPath + "bricks/")
] + [
    os.path.join(isolationsPath, "facades/", x)
    for x in os.listdir(isolationsPath + "facades/")
]
fonts = os.listdir(isolationsPath + "fonts/")


def printProgressBar(iteration, total, prefix="", suffix="", decimals=3, fill="█"):
    if os.popen("stty size", "r").read():
        columns = int(os.popen("stty size", "r").read().split()[1])
        length = columns - 9 - decimals
        percent = ("{0:." + str(decimals) + "f}").format(
            100 * (iteration / float(total))
        )
        filledLength = int(length * iteration // total)
        bar = fill * filledLength + "-" * (length - filledLength)
        print("\r%s |%s| %s%% %s" % (prefix, bar, percent, suffix), end="\r")


def pickRandomLineFromFile(filename):
    with open(filename, "r") as f:
        lines = f.readlines()
        return random.choice(lines)


def generateRandomColor(style):
    rgb = None
    if style is "dark":
        rgb = [random.randint(0, 80) for i in range(3)]
    elif style is "light":
        rgb = [random.randint(175, 255) for i in range(3)]
    else:
        rgb = [random.randint(0, 255) for i in range(3)]
    rgb = [hex(i)[2:].zfill(2) for i in rgb]
    return ("#" + ''.join(rgb)).upper()


def generateTextImage(isolations, modifications, debug=False):
    height = random.randint(500, 800)
    width = random.randint(800, 1200)

    textImage = Image.new("RGB", (width, height), generateRandomColor("dark"))

    fontPath = os.path.join(isolationsPath, "fonts", random.choice(fonts))
    fontSize = random.randint(300, 400)
    textPadding = 30
    font, textToDraw, textSize, textCoords = None, None, None, None
    while True:
        font = ImageFont.truetype(fontPath, fontSize)
        textToDraw = str(random.randint(0, 3000))
        textSize = ImageDraw.ImageDraw(
            textImage).textsize(textToDraw, font=font)
        textCoords = (((width - textSize[0]) / 2),
                      ((height - textSize[1]) / 2))
        if textCoords[0] + textSize[0] > width - textPadding or textCoords[1] + textSize[1] > height - textPadding:
            if fontSize == 0:
                break
            fontSize -= 5
            continue
        else:
            break
    if fontSize == 0:
        return None

    color = generateRandomColor("light")
    ImageDraw.Draw(textImage).text(
        textCoords, textToDraw, fill=color, font=font)
    if debug:
        ImageDraw.Draw(textImage).rectangle(
            xy=[textCoords, tuple([sum(x)
                                   for x in zip(textCoords, textSize)])],
            outline="green", width=10)

    textImage = np.array(textImage)
    textImage = cv2.cvtColor(
        textImage, cv2.COLOR_BGR2RGB)  # Convert RGB to BGR
    
    seq = iaa.Sequential(
        [
            iaa.AdditiveGaussianNoise(loc=0, scale=(1, 10)),
            iaa.Sometimes(0.5, iaa.GaussianBlur(sigma=(1.0, 2.0))),
            iaa.Dropout(p=(0, 0.05), per_channel=0.5),
            iaa.ElasticTransformation(alpha=(0, 2.0), sigma=0.25),
            iaa.AddElementwise((-20, 20)),
            iaa.AddToHueAndSaturation((-30, 0))
        ]
    ).to_deterministic()

    textImage = seq.augment_image(textImage)

    x1 = textCoords[0]
    x2 = textCoords[0] + textSize[0]
    y1 = textCoords[1]
    y2 = textCoords[1] + textSize[1]

    numberCoordsBefore = np.float32([[x1, y1], [x1, y2], [x2, y1], [x2, y2]])
    numberCoordsAfter = np.float32(
        [[j + random.randrange(-80, 80) for j in i]
         for i in numberCoordsBefore]
    )

    M = cv2.getPerspectiveTransform(numberCoordsBefore, numberCoordsAfter)

    rows, cols, ch = textImage.shape
    textImage = cv2.warpPerspective(textImage, M, (cols, rows))

    cv2.imwrite("sample.png", textImage)
    return (textImage, height, width, numberCoordsAfter, textToDraw)


def generateBackground(isolations, modifications, debug=False):
    height = 10 * int(pickRandomLineFromFile("heights.txt")) + 5
    width = 10 * int(pickRandomLineFromFile("widths.txt")) + 5

    background = Image.open(random.choice(backgrounds))
    background = ImageOps.fit(background, (width, height), Image.NEAREST)
    background = np.array(background)
    background = cv2.cvtColor(
        background, cv2.COLOR_BGR2RGB)  # Convert RGB to BGR

    # We want our maximum rotation angle to be x/10+1, where x is the
    # ratio of the image's width to its height. Why? Because I said so
    maxRotation = (width / height) / 10 + 1

    seq = iaa.Sequential(
        [
            iaa.Sometimes(0.25, iaa.Superpixels(p_replace=0.4, n_segments=64)),
            iaa.Multiply((0.8, 1.2), per_channel=0.15),
            iaa.AdditiveGaussianNoise(loc=0, scale=(1, 10)),
            iaa.Sometimes(0.5, iaa.GaussianBlur(sigma=(1.0, 2.0))),
            iaa.Affine(rotate=(-maxRotation, maxRotation), shear=(-3, 3)),
            iaa.PiecewiseAffine(scale=(0.005, 0.03)),
            iaa.Fliplr(0.5),
            iaa.Flipud(0.25),
            iaa.Dropout(p=(0, 0.05), per_channel=0.5),
            iaa.ElasticTransformation(alpha=(0, 2.0), sigma=0.25),
            # iaa.ContrastNormalization((0.5, 1.5)),
            # iaa.Sharpen(alpha=(0.0, 0.5)),
            iaa.AddElementwise((-20, 20)),
            iaa.AddToHueAndSaturation((-30, 0))
        ]
    ).to_deterministic()

    background = seq.augment_image(background)

    bgCoordsBefore = np.float32(
        [[0, 0], [0, height], [width, 0], [width, height]])
    bgCoordsAfter = np.float32(
        [[j + random.uniform(-width / 10, width / 10)
          for j in i] for i in bgCoordsBefore]
    )
    M = cv2.getPerspectiveTransform(bgCoordsBefore, bgCoordsAfter)

    rows, cols, ch = background.shape
    background = cv2.warpPerspective(background, M, (cols, rows))

    return (background, height, width)


def generateImage(isolations, modifications, debug=False):
    (textImage, textImageHeight, textImageWidth, numberCoordinates,
     houseNumber) = generateTextImage(isolations, modifications, debug)
    (backgroundImage, backgroundHeight, backgroundWidth) = generateBackground(
        isolations, modifications, debug)

    textImageInverseScale = max(
        textImageHeight / backgroundHeight, textImageWidth / backgroundWidth)
    c = tuple([round(x / textImageInverseScale)
               for x in [textImageWidth, textImageHeight]])
    textImage = cv2.resize(textImage, c)
    numberCoordinates = [x / textImageInverseScale for x in numberCoordinates]

    textImageHeight = textImage.shape[0]
    textImageWidth = textImage.shape[1]

    x_offset = (backgroundWidth - textImageWidth) // 2
    y_offset = (backgroundHeight - textImageHeight) // 2
    y1, y2 = y_offset, y_offset + textImage.shape[0]
    x1, x2 = x_offset, x_offset + textImage.shape[1]

    numberCoordinates = [[(i[0] + x_offset) / 10, (i[1] + y_offset) / 10]
                         for i in numberCoordinates]

    # alpha = cv2.cvtColor(textImage, cv2.COLOR_BGR2GRAY)
    r, alpha = cv2.threshold(textImage, 0, 1, cv2.THRESH_BINARY)

    backgroundImage[y1:y2, x1:x2] = backgroundImage[y1:y2,
                                                    x1:x2] * (1.3-alpha) + textImage * alpha * 0.7 

    backgroundImage = cv2.resize(
        backgroundImage, (backgroundWidth // 10, backgroundHeight // 10))

    if debug:
        verticalPairs = writeBoundingBoxes(
            "", "", numberCoordinates, houseNumber, shouldSave=False
        )
        verticalPairs = np.float32(verticalPairs).reshape(
            (-1, 1, 2)).astype("int32")
        backgroundImage = cv2.polylines(
            backgroundImage, verticalPairs, True, (25, 0, 255), 4)

    return(backgroundImage, numberCoordinates, houseNumber)


def writeBoundingBoxes(filename, imagename, coords, streetNumber, shouldSave=True):
    n = len(streetNumber)
    xa1 = coords[0][0]
    ya1 = coords[0][1]
    xa2 = coords[2][0]
    ya2 = coords[2][1]

    xb1 = coords[1][0]
    yb1 = coords[1][1]
    xb2 = coords[3][0]
    yb2 = coords[3][1]

    verticalPairs = [
        [
            [xa1 + (xa2 - xa1) / n * i, ya1 + (ya2 - ya1) / n * i],
            [xb1 + (xb2 - xb1) / n * i, yb1 + (yb2 - yb1) / n * i],
        ]
        for i in range(0, n + 1)
    ]

    if shouldSave:
        with open(filename, "a") as csv:
            for index in range(0, len(verticalPairs) - 1):
                p1 = verticalPairs[index][0]
                p2 = verticalPairs[index + 1][1]
                csv.write(
                    "{},{},{},{},{}\n".format(
                        imagename, p1[0], p1[1], p2[0], p2[1]
                    )
                )
    return verticalPairs


def main():
    parser = argparse.ArgumentParser(
        description="Build a synthetic SVHN-esque dataset."
    )
    parser.add_argument(
        "--count", type=int, default=20000, help="The number of images created."
    )
    parser.add_argument(
        "--isolations",
        choices=["low", "medium", "high"],
        nargs=1,
        # required=True,
        help="The quantity of isolations used when generating data.",
    )
    parser.add_argument(
        "--modifications",
        choices=["low", "medium", "high"],
        nargs=1,
        # required=True,
        help="The amount of modification used when generating data.",
    )
    parser.add_argument(
        "--outdir",
        default="img_out/",
        help="The directory in which output images should be stored.",
    )
    parser.add_argument(
        "--cleardir",
        action="store_true",
        help="Should the output directory be cleared before images are generated?",
    )
    parser.add_argument(
        "--profile",
        action="store_true",
        help="Should the script output profiling information?",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Should debugging dots be drawn on the images?",
    )
    args = parser.parse_args()

    pr = cProfile.Profile()
    if args.profile:
        pr.enable()

    print("Getting things ready...")
    if not os.path.exists(args.outdir):
        os.makedirs(args.outdir)

    csvFilename = args.outdir + ".csv"

    if args.cleardir:
        open(csvFilename, "w").close()
        for f in os.listdir(args.outdir):
            os.remove(os.path.join(args.outdir, f))

    printProgressBar(0, args.count)
    for i in range(0, args.count):
        (image, outputCoordinates, outputNumber) = generateImage(
            args.isolations, args.modifications, args.debug
        )
        filename = os.path.join(args.outdir, "{}.jpg".format(i + 1))
        cv2.imwrite(filename, image)
        writeBoundingBoxes(csvFilename, filename,
                           outputCoordinates, outputNumber)
        printProgressBar(i, args.count)
    printProgressBar(args.count, args.count)
    print("Done!")

    if args.profile:
        pr.disable()
        sortby = "cumulative"
        ps = pstats.Stats(pr).sort_stats(sortby)
        ps.print_stats()


main()
