#!/bin/python3
# ./datagen.py --cleardir --count 80

import os, glob, random, argparse
from PIL import Image, ImageOps
from functools import reduce

def printProgressBar(iteration, total, prefix="", suffix="", decimals=3, fill="█"):
    if os.popen("stty size", "r").read():
        columns = int(os.popen("stty size", "r").read().split()[1])
        length = columns - 9 - decimals
        percent = ("{0:." + str(decimals) + "f}").format(
            100 * (iteration / float(total))
        )
        filledLength = int(length * iteration // total)
        bar = fill * filledLength + "-" * (length - filledLength)
        print("\r%s |%s| %s%% %s" % (prefix, bar, percent, suffix), end="\r")

class BBox:
    def __init__(self, x, y, width, height):
        self.x = x
        self.y = y
        self.width = width
        self.height = height


class AnnotatedImage:
    def __init__(self, image, bboxList):
        self.image = image
        self.bboxList = bboxList

    def saveImage(self, filename):
        self.image.save(filename)

    def writeBBoxes(self, imageFilename, csvFilename):
        with open(csvFilename, "a") as csv:
            for bbox in self.bboxList:
                csv.write("{},{},{},{},{}\n".format(
                    imageFilename,
                    bbox.x,
                    bbox.y,
                    bbox.x + bbox.width,
                    bbox.y + bbox.height
                ))
        return len(self.bboxList)

isolationsPath = "../isolations"
opj = os.path.join

backgroundPaths = glob.glob(opj(isolationsPath, "bg*.jpg"))
digitPaths = glob.glob(opj(isolationsPath, "digit*.jpg"))

if len(backgroundPaths) is 0 or len(digitPaths) is 0:
    exit()

def imageGen():
    # Pick a random background
    backgroundPath = random.choice(backgroundPaths)
    background = Image.open(backgroundPath).convert("RGBA")

    # Pick a quantity of digits
    digitQuantity = random.choice([1, 2, 2, 3, 3, 3, 3, 4, 4])  # hacky distribution

    # Create an empty digit images list
    digitImages = []

    # For i in quantity:
    for i in range(0, digitQuantity):
        # Pick a random image
        digitPath = random.choice(digitPaths)
        digit = Image.open(digitPath)
        digitValue = digitPath.split("-")[1].split(".")[0]  # "digit12-6.jpg" -> "6"

        # Scale it
        heightRatio = random.random() * 0.55 + 0.3 # Random float between 0.3 and 0.85
        outputHeight = int(round(background.height * heightRatio))
        outputWidth = int(round(outputHeight / digit.height * digit.width))
        digit = ImageOps.fit(digit, (outputWidth, outputHeight))

        # Rotate it
        digit = digit.convert("RGBA").rotate(random.random() * 20 - 10, Image.BICUBIC, expand=1)

        # Append to digit images list
        digitImages.append(digit)

    # Join all images together horizontally

    # Get the output width of allDigits plate
    totalWidth = 0
    for image in digitImages:
        totalWidth += image.width

    # Generate a total height accounting for jitter
    totalHeight = max([x.height for x in digitImages])
    heightJitter = int(round(totalHeight * 0.3))
    totalHeight += heightJitter

    # Composite image of all digits
    allDigits = Image.new("RGBA", (totalWidth, totalHeight), (0, 0, 0, 0))
    bboxList = []  # List of all bounding boxes

    xOffset = 0
    for image in digitImages:
        # Calculate y offset
        yOffset = int(round((totalHeight - image.height) / 2))
        jitter = int(round((random.random() * heightJitter * 2))) - heightJitter # Random float between -jitter and +jitter
        yOffset += jitter

        # Put image on composite
        # print(allDigits)
        allDigits.paste(image, (xOffset, yOffset), image)

        # Initial x/y is offset by paste coordinate but we'll correct that later
        bbox = BBox(xOffset, yOffset, image.width, image.height)
        bboxList.append(bbox)

        xOffset += image.width

    # Place that image in the center of the background
    pasteCoordinateWidth = int(round(((background.width - allDigits.width) / 2)))
    pasteCoordinateHeight = int(round(((background.height - allDigits.height) / 2)))
    pasteCoordinate = (pasteCoordinateWidth, pasteCoordinateHeight)

    background.paste(allDigits, pasteCoordinate, allDigits)
    background = background.convert("RGB")

    for bbox in bboxList:
        bbox.x += pasteCoordinateWidth
        bbox.y += pasteCoordinateHeight

    # Generate AnnotatedImage object and return
    return AnnotatedImage(background, bboxList)


parser = argparse.ArgumentParser(
    description="Build a synthetic SVHN-esque dataset."
)

parser.add_argument(
    "--count", type=int, default=20000, help="The number of images generated."
)

parser.add_argument(
    "--outdir",
    default="../output/",
    help="The directory in which output images should be stored.",
)

parser.add_argument(
    "--cleardir",
    action="store_true",
    help="Clear the output directory before images are generated",
)

args = parser.parse_args()

print("Getting things ready...")
if not os.path.exists(args.outdir):
    os.makedirs(args.outdir)

csvFilename = opj(args.outdir, "bboxes.csv")

if args.cleardir:
    open(csvFilename, "w").close()
    for f in os.listdir(args.outdir):
        os.remove(opj(args.outdir, f))

printProgressBar(0, args.count)
bboxesCount = 0
for i in range(0, args.count):
    imageFilename = opj(args.outdir, "{}.jpg".format(i + 1))
    annotatedImage = imageGen()
    annotatedImage.saveImage(imageFilename)
    bboxesCount += annotatedImage.writeBBoxes(imageFilename, csvFilename)
    printProgressBar(i, args.count)
printProgressBar(args.count, args.count)
print("Done. {} images generated, {} BBoxes written.".format(
    args.count, bboxesCount
))
